REBOL [
	Title: "REBOL/Services - Password Hashing"
	Author: "Carl Sassenrath"
	Notice: "Confidential source code. Under development."
	Copyright: "REBOL Technology 2005-2008, All rights reserved."
]

;-- Password Hashing Functions -----------------------------------------------
;
;	This password system uses a highly secure SHA1 hashing algorithm for
;	both password storage and authentication. The actual password is not
;	stored nor sent over the network nor revealed in any manner.
;
;	Authentication is done with these steps:
;		1. Client sends username to service.
;		2. Service looks up user, sends salted challenge and the salt.
;		3. Client (if necessary) hashes pass string with the salt.
;		4. Client responds by hashing hashed pass with challenge. 
;		5. Server authenticates by comparing with the hashed result.
;
;	For security, probing for valid user names will fail. If a user does
;	not exist, the system still returns a challenge, not an error.

random/seed now

make-salt: does [
	checksum/secure form random/secure to-integer #{7fffffff}
]

make-pass: func [
	"Returns a binary password hash. Case sensitive."
	pass
	salt
][
	checksum/secure append to-binary pass salt
]

pass-challenge: func [
	"Returns a password challenge hash."
	pass
	challenge
][
	checksum/secure make-pass pass challenge
]
