REBOL [
	Title: "REBOL/Services - Home Service Dialect"
	Author: "Carl Sassenrath"
	Notice: "Confidential source code. Under development."
	Copyright: "REBOL Technology 2005-2008, All rights reserved."
]

;   [service name | none] -> [service name | none]
;   [options opt-list | none] -> [options opt-list | none]
;   [echo any-values] -> [echo any-values]
;   [info fields | all | none] -> [info name1 value1 name2 value2 ...]

add-default-service [
	name: 'home
	title: "Home Service"
	description: "Implements the home service (it is always available)."
	language: 'en
	commands: [

		;-- System Commands:
		'service #doc [
				en "Change the service command context" 
				it "Cambia il contesto dei comandi"
			]
			set args opt word! "Name of service" (
				if not set-service session args [bad-service args]
				result: next command
			)

		;-- User Account Authentication:
		|
			'login #doc [
				en "Login into a user account (force authentication)"
				it "Accedi come utente specifico (forza autenticazione)"
			]
			args: string! "Username" opt integer! "Encrypt strength" (
				result: login-session session args/1 args/2
				if word? result [throw result]
			)
		|
			'logout #doc [
				en "Logout out of a user account"
				it "Termina l'accesso come utente specifico"
			] (
			if not session/login-date [throw 'no-login]
			result: logout-session session
		)

		;-- Utility Commands:
		|
			'echo #doc [
				en "Echo whatever values were passed" 
				it "Ripete qualunque valore inviato"
			]
			result: to end
		|
			'client #doc [
				en "Download the standard service client"
				it "Scarica il client standard"
			]
			[
				'version (result: session/config/service-info/client-version)
				|
				none (get default client code)
			]
		|
			'end #doc [
				en "Terminate session (for CGI efficiency)"
				it "Termina la sessione (per efficienza nel caso CGI)"
			]

		;-- Information Commands:
		|
			'info #doc [
				en "Get general info about the service server"
				it "Riporta informazioni generali riguardo il server"
			]
			args: ['all | any word! | none] "Optional info field names"
			(result: reply-info session args)
		|
			'commands #doc [
				en "Return command summary for this service"
				it "Ritorna un sommario dei comandi per questo servizio"
			]
			(result: make-service-doc commands any [session/language language])
		|
			'date #doc [
				en "Return the server's current date/time/zone"
				it "Ritorna la data/ora/fuso orario corrente del server"
			]
			(result: now)
		|
			'system #doc [
				en "Get REBOL system product and version info"
				it "Riporta informazioni sul tipo e la versione del sistema REBOL"
			]
			(result: reduce ['REBOL system/product system/version system/build])
		|
			'language #doc [
				en "Set client's target language (EN FR ...) [session]"
				it "Imposta la lingua del client (IT EN ...) [sessione]"
			]
			set args opt word! "Language id" (
				if args [session/language: args]
				result: any [session/language language]
			)
		|
			'session
			(result: mold next first session)
		|
			'quit (
				close-all-sessions session/config
				close serv-port
				print "Quit..."
				wait 1
				quit
			)

		;-- Feedback Channels:
		|
			'feedback #doc [
				en "Send feedback to the service operator"
				it "Invia del feedback all'operatore del servizio"
			]
			(result: 'not-implemented)
		|
			'invite #doc [
				en "Ask for an account on this service"
				it "Richiede un account per questo servizio"
			]
			(result: 'not-implemented)

		;-- Miscellaneous:
		|
			'change-pass #doc [
				en "Change a user's password."
				it "Cambia la password di un utente."
			]
			set args [string! | binary!]
			(
				if session/user-id = 0 [throw 'login-required]
				change-pass session/user-id args
				save-userbase
				result: true
			)
		|
			'user-id "Return valid user id if session is logged in"
			(result: session/user-id)
		|
			'options "Pending - not implemented"
			to end
			(result: next command)
		|
			'fatal "Test option - ignore"
			(throw fatal error - is that possible? for testing)
	]

	reply-info: func [
		"Parse the INFO command and construct requested fields in the reply:"
		session
		command
		/local words result
	][
		words: any [
			; Requesting ALL info?
			if find command 'all [next first session/config/service-info]
			; Requesting specific words?
			if not tail? command [command]
			; Requesting default info?
			[title summary version]
		]
		; Fetch the above field values:
		result: copy []
		foreach word words [
			if word? word [
				append result word
				append/only result get in session/config/service-info word
			]
		]
		result
	]
]
